@@ -5,20 +5,25 @@ mod tests;
 use std::borrow::Cow;
 use std::path::PathBuf;
 use std::str::FromStr;
-use std::sync::{Arc, Mutex};
+use std::sync::{Arc, Mutex, OnceLock};
 use std::time::Instant;
-use std::{fs, io};
+use std::{
+    fs,
+    io::{self, Read},
+};
 
 use crate::debug_impls::{DebugBytesPrefix, DebugInline};
 use crate::interpreter::ImageCallback;
+use crate::metrics::{histogram, HistTag};
 use crate::positioner::DEFAULT_MARGIN;
-use crate::utils::{self, usize_in_mib, Align, Point, Size};
+use crate::utils::{usize_in_mib, Align, Point, Size};
 
 use anyhow::Context;
 use bytemuck::{Pod, Zeroable};
 use image::{ImageBuffer, RgbaImage};
+use resvg::{tiny_skia, usvg};
 use smart_debug::SmartDebug;
-use usvg::{TreeParsing, TreeTextToPath};
+use usvg::fontdb;
 use wgpu::util::DeviceExt;
 use wgpu::{BindGroup, Device, TextureFormat};
 
@@ -40,7 +45,7 @@ impl From<u32> for Px {
     }
 }
 
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, PartialEq)]
 pub enum ImageSize {
     PxWidth(Px),
     PxHeight(Px),
@@ -56,7 +61,7 @@ impl ImageSize {
     }
 }
 
-#[derive(SmartDebug, Default, Clone)]
+#[derive(SmartDebug, Default, Clone, PartialEq)]
 pub struct ImageData {
     #[debug(wrapper = DebugBytesPrefix)]
     lz4_blob: Vec<u8>,
@@ -66,7 +71,7 @@ pub struct ImageData {
 }
 
 impl ImageData {
-    fn load(bytes: &[u8], scale: bool) -> anyhow::Result<Self> {
+    pub fn load(bytes: &[u8], scale: bool) -> anyhow::Result<Self> {
         let (lz4_blob, dimensions) = decode::decode_and_compress(bytes)?;
         Ok(Self {
             lz4_blob,
@@ -108,6 +113,8 @@ impl ImageData {
 
 #[derive(SmartDebug, Default)]
 pub struct Image {
+    // TODO: Instead of sharing a mutex with the image loading thread change this to hold a oneshot
+    // channel that stores the image?
     #[debug(skip_fn = debug_ignore_image_data)]
     pub image_data: Arc<Mutex<Option<ImageData>>>,
     #[debug(skip_fn = Option::is_none, wrapper = DebugInline)]
@@ -122,6 +129,49 @@ pub struct Image {
     pub hidpi_scale: f32,
 }
 
+// NOTE: Internally performs some expensive operations. Avoid calling often
+// TODO: can we deprecate this and then `allow` specific usages of it due to ^^?
+impl PartialEq for Image {
+    fn eq(&self, other: &Self) -> bool {
+        let Self {
+            image_data,
+            is_aligned,
+            size,
+            bind_group,
+            is_link,
+            hidpi_scale,
+        } = self;
+        let Self {
+            image_data: other_image_data,
+            is_aligned: other_is_aligned,
+            size: other_size,
+            bind_group: other_bind_group,
+            is_link: other_is_link,
+            hidpi_scale: other_hidpi_scale,
+        } = other;
+
+        let clone_image_data = |shared_image: &Mutex<Option<_>>| {
+            shared_image
+                .lock()
+                .unwrap_or_else(|err| err.into_inner())
+                .to_owned()
+        };
+        let image_data = clone_image_data(image_data);
+        let other_image_data = clone_image_data(other_image_data);
+
+        let some_bind_group = bind_group.is_some();
+        let some_other_bind_group = other_bind_group.is_some();
+        let bind_group_variant_matches = some_bind_group ^ !some_other_bind_group;
+
+        image_data == other_image_data
+            && is_aligned == other_is_aligned
+            && size == other_size
+            && is_link == other_is_link
+            && hidpi_scale == other_hidpi_scale
+            && bind_group_variant_matches
+    }
+}
+
 fn debug_ignore_image_data(mutex: &Mutex<Option<ImageData>>) -> bool {
     match mutex.lock() {
         Ok(data) => data.is_none(),
@@ -217,6 +267,8 @@ impl Image {
         let image_data_clone = image_data.clone();
 
         std::thread::spawn(move || {
+            let start = Instant::now();
+
             let mut src_path = PathBuf::from(&src);
             if src_path.is_relative() {
                 if let Some(parent_dir) = file_path.parent() {
@@ -226,12 +278,8 @@ impl Image {
 
             let image_data = if let Ok(img_file) = fs::read(&src_path) {
                 img_file
-            } else if let Ok(bytes) = utils::client()
-                .get(&src)
-                .send()
-                .and_then(|resp| resp.bytes())
-            {
-                bytes.to_vec()
+            } else if let Ok(bytes) = http_get_image(&src) {
+                bytes
             } else {
                 tracing::warn!("Request for image from {} failed", src_path.display());
                 return;
@@ -241,8 +289,6 @@ impl Image {
                 image
             } else {
                 let opt = usvg::Options::default();
-                let mut fontdb = usvg::fontdb::Database::new();
-                fontdb.load_system_fonts();
                 // TODO: yes all of this image loading is very messy and could use a refactor
                 let Ok(mut tree) = usvg::Tree::from_data(&image_data, &opt) else {
                     tracing::warn!(
@@ -264,13 +310,18 @@ impl Image {
                     )
                     .unwrap(),
                 );
-                tree.convert_text(&fontdb);
-                let rtree = resvg::Tree::from_usvg(&tree);
+                static FONTDB: OnceLock<fontdb::Database> = OnceLock::new();
+                let fontdb = FONTDB.get_or_init(|| {
+                    let mut db = fontdb::Database::new();
+                    db.load_system_fonts();
+                    db
+                });
+                tree.postprocess(Default::default(), fontdb);
                 let mut pixmap =
-                    tiny_skia::Pixmap::new(rtree.size.width() as u32, rtree.size.height() as u32)
+                    tiny_skia::Pixmap::new(tree.size.width() as u32, tree.size.height() as u32)
                         .context("Couldn't create svg pixmap")
                         .unwrap();
-                rtree.render(tiny_skia::Transform::default(), &mut pixmap.as_mut());
+                resvg::render(&tree, tiny_skia::Transform::default(), &mut pixmap.as_mut());
                 ImageData::new(
                     ImageBuffer::from_raw(pixmap.width(), pixmap.height(), pixmap.data().into())
                         .context("Svg buffer has invalid dimensions")
@@ -280,6 +331,7 @@ impl Image {
             };
 
             *image_data_clone.lock().unwrap() = Some(image);
+            histogram!(HistTag::ImageLoad).record(start.elapsed());
             image_callback.loaded_image(src, image_data_clone);
         });
 
@@ -375,6 +427,26 @@ impl Image {
     }
 }
 
+pub fn http_get_image(url: &str) -> anyhow::Result<Vec<u8>> {
+    const USER_AGENT: &str = concat!(
+        "inlyne ",
+        env!("CARGO_PKG_VERSION"),
+        " https://github.com/Inlyne-Project/inlyne"
+    );
+
+    const LIMIT: usize = 20 * 1_024 * 1_024;
+
+    let resp = ureq::get(url).set("User-Agent", USER_AGENT).call()?;
+    let len = resp
+        .header("Content-Length")
+        .and_then(|len| len.parse::<usize>().ok());
+    let mut body = Vec::with_capacity(len.unwrap_or(0).clamp(0, LIMIT));
+    resp.into_reader()
+        .take(u64::try_from(LIMIT).unwrap())
+        .read_to_end(&mut body)?;
+    Ok(body)
+}
+
 #[repr(C)]
 #[derive(Clone, Copy, Pod, Zeroable, Debug)]
 pub struct ImageVertex {
