@@ -1,5 +1,6 @@
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
 pub enum Action {
+    History(HistDirection),
     ToEdge(VertDirection),
     Scroll(VertDirection),
     Page(VertDirection),
@@ -8,6 +9,12 @@ pub enum Action {
     Quit,
 }
 
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+pub enum HistDirection {
+    Next,
+    Prev,
+}
+
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
 pub enum VertDirection {
     Up,
