@@ -14,13 +14,18 @@ pub mod color;
 mod debug_impls;
 mod file_watcher;
 pub mod fonts;
+pub mod history;
 pub mod image;
 pub mod interpreter;
 mod keybindings;
+mod metrics;
 pub mod opts;
+mod panic_hook;
 pub mod positioner;
 pub mod renderer;
+pub mod selection;
 pub mod table;
+#[cfg(test)]
 pub mod test_utils;
 pub mod text;
 pub mod utils;
@@ -34,13 +39,15 @@ use std::str::FromStr;
 use std::sync::atomic::{AtomicBool, Ordering};
 use std::sync::mpsc::{self, channel};
 use std::sync::{Arc, Mutex};
+use std::time::Instant;
 
 use file_watcher::Watcher;
 use image::{Image, ImageData};
 use interpreter::HtmlInterpreter;
-use keybindings::action::{Action, VertDirection, Zoom};
+use keybindings::action::{Action, HistDirection, VertDirection, Zoom};
 use keybindings::{Key, KeyCombos, ModifiedKey};
-use opts::{Args, Config, Opts};
+use metrics::{histogram, HistTag};
+use opts::{Cli, Config, Opts};
 use positioner::{Positioned, Row, Section, Spacer, DEFAULT_MARGIN, DEFAULT_PADDING};
 use raw_window_handle::HasRawDisplayHandle;
 use renderer::Renderer;
@@ -50,13 +57,16 @@ use tracing_subscriber::prelude::*;
 use tracing_subscriber::util::SubscriberInitExt;
 use utils::{ImageCache, Point, Rect, Size};
 
+use crate::opts::{Commands, ConfigCmd, MetricsExporter};
+use crate::selection::Selection;
 use anyhow::Context;
+use clap::Parser;
 use taffy::Taffy;
 use winit::event::{
     ElementState, Event, KeyboardInput, ModifiersState, MouseButton, MouseScrollDelta, WindowEvent,
 };
 use winit::event_loop::{ControlFlow, EventLoop, EventLoopBuilder};
-use winit::window::{CursorIcon, Window};
+use winit::window::{CursorIcon, Window, WindowBuilder};
 
 pub enum InlyneEvent {
     LoadedImage(String, Arc<Mutex<Option<ImageData>>>),
@@ -78,7 +88,7 @@ pub enum Hoverable<'a> {
     Summary(&'a Section),
 }
 
-#[derive(Debug)]
+#[derive(Debug, PartialEq)]
 pub enum Element {
     TextBox(TextBox),
     Spacer(Spacer),
@@ -140,51 +150,35 @@ pub struct Inlyne {
     keycombos: KeyCombos,
     need_repositioning: bool,
     watcher: Watcher,
-}
-
-/// Gets a relative path extending from the repo root falling back to the full path
-fn root_filepath_to_vcs_dir(path: &Path) -> Option<PathBuf> {
-    let mut full_path = path.canonicalize().ok()?;
-    let mut parts = vec![full_path.file_name()?.to_owned()];
-
-    full_path.pop();
-    loop {
-        full_path.push(".git");
-        let is_git = full_path.exists();
-        full_path.pop();
-        full_path.push(".hg");
-        let is_mercurial = full_path.exists();
-        full_path.pop();
-
-        let is_vcs_dir = is_git || is_mercurial;
-
-        match full_path.file_name() {
-            Some(name) => parts.push(name.to_owned()),
-            // We've searched the full path and didn't find a vcs dir
-            None => return Some(path.to_owned()),
-        }
-        if is_vcs_dir {
-            let mut rooted = PathBuf::new();
-            for part in parts.into_iter().rev() {
-                rooted.push(part);
-            }
-            return Some(rooted);
-        }
-
-        full_path.pop();
-    }
+    selection: Selection,
 }
 
 impl Inlyne {
     pub fn new(opts: Opts) -> anyhow::Result<Self> {
         let keycombos = KeyCombos::new(opts.keybindings.clone())?;
 
+        let file_path = opts.history.get_path().to_owned();
+
         let event_loop = EventLoopBuilder::<InlyneEvent>::with_user_event().build();
-        let window = Arc::new(Window::new(&event_loop).unwrap());
-        match root_filepath_to_vcs_dir(&opts.file_path) {
-            Some(path) => window.set_title(&format!("Inlyne - {}", path.to_string_lossy())),
-            None => window.set_title("Inlyne"),
-        }
+
+        let window = {
+            let mut wb = WindowBuilder::new().with_title(utils::format_title(&file_path));
+
+            if let Some(ref pos) = opts.position {
+                wb = wb.with_position(winit::dpi::PhysicalPosition::new(pos.x, pos.y));
+            }
+            if let Some(ref size) = opts.size {
+                wb = wb.with_inner_size(winit::dpi::PhysicalSize::new(size.width, size.height));
+            }
+            #[cfg(all(feature = "wayland", not(any(target_os = "macos", windows))))]
+            {
+                use winit::platform::wayland::WindowBuilderExtWayland;
+                wb = wb.with_name("inlyne", "");
+            }
+
+            Arc::new(wb.build(&event_loop).unwrap())
+        };
+
         let renderer = pollster::block_on(Renderer::new(
             &window,
             opts.theme.clone(),
@@ -195,8 +189,8 @@ impl Inlyne {
 
         let element_queue = Arc::new(Mutex::new(VecDeque::new()));
         let image_cache = Arc::new(Mutex::new(HashMap::new()));
-        let md_string = read_to_string(&opts.file_path)
-            .with_context(|| format!("Could not read file at '{}'", opts.file_path.display()))?;
+        let md_string = read_to_string(&file_path)
+            .with_context(|| format!("Could not read file at '{}'", file_path.display()))?;
 
         let interpreter = HtmlInterpreter::new(
             window.clone(),
@@ -204,7 +198,7 @@ impl Inlyne {
             renderer.theme.clone(),
             renderer.surface_format,
             renderer.hidpi_scale,
-            opts.file_path.clone(),
+            file_path.clone(),
             image_cache.clone(),
             event_loop.create_proxy(),
             opts.color_scheme,
@@ -218,7 +212,9 @@ impl Inlyne {
 
         let lines_to_scroll = opts.lines_to_scroll;
 
-        let watcher = Watcher::spawn(event_loop.create_proxy(), opts.file_path.clone());
+        let watcher = Watcher::spawn(event_loop.create_proxy(), file_path.clone());
+
+        let _ = file_path.parent().map(std::env::set_current_dir);
 
         Ok(Self {
             opts,
@@ -234,6 +230,7 @@ impl Inlyne {
             keycombos,
             need_repositioning: false,
             watcher,
+            selection: Selection::new(),
         })
     }
 
@@ -248,6 +245,8 @@ impl Inlyne {
                 .map(|mut queue| queue.drain(..).collect::<Vec<Element>>())
         };
         if let Ok(queue) = queue {
+            let positioning_start = Instant::now();
+
             for element in queue {
                 // Position element and add it to elements
                 let mut positioned_element = Positioned::new(element);
@@ -264,6 +263,8 @@ impl Inlyne {
                         + positioned_element.bounds.as_ref().unwrap().size.1;
                 elements.push(positioned_element);
             }
+
+            histogram!(HistTag::Positioner).record(positioning_start.elapsed());
         }
     }
 
@@ -278,14 +279,18 @@ impl Inlyne {
         self.interpreter_sender.send(contents).unwrap();
     }
 
+    fn update_file(&mut self, path: &Path, contents: String) {
+        self.window.set_title(&utils::format_title(path));
+        self.watcher.update_file(path, contents);
+        self.renderer.set_scroll_y(0.0);
+    }
+
     pub fn run(mut self) {
         let mut pending_resize = None;
         let mut scrollbar_held = None;
         let mut mouse_down = false;
         let mut modifiers = ModifiersState::empty();
-        let mut last_loc = (0.0, 0.0);
-        let mut selection_cache = String::new();
-        let mut selecting = false;
+        let mut mouse_position: Point = Point::default();
 
         let event_loop = self.event_loop.take().unwrap();
         let event_loop_proxy = event_loop.create_proxy();
@@ -301,12 +306,12 @@ impl Inlyne {
                         self.image_cache.lock().unwrap().insert(src, image_data);
                         self.need_repositioning = true;
                     }
-                    InlyneEvent::FileReload => match read_to_string(&self.opts.file_path) {
+                    InlyneEvent::FileReload => match read_to_string(self.opts.history.get_path()) {
                         Ok(contents) => self.load_file(contents),
                         Err(err) => {
                             tracing::warn!(
                                 "Failed reloading file at {}\nError: {}",
-                                self.opts.file_path.display(),
+                                self.opts.history.get_path().display(),
                                 err
                             );
                         }
@@ -325,6 +330,7 @@ impl Inlyne {
                     }
                 },
                 Event::RedrawRequested(_) => {
+                    let redraw_start = Instant::now();
                     Self::position_queued_elements(
                         &self.element_queue,
                         &mut self.renderer,
@@ -332,12 +338,11 @@ impl Inlyne {
                     );
                     self.renderer.set_scroll_y(self.renderer.scroll_y);
                     self.renderer
-                        .redraw(&mut self.elements)
+                        .redraw(&mut self.elements, &mut self.selection)
                         .context("Renderer failed to redraw the screen")
                         .unwrap();
-                    if selecting {
-                        selection_cache = self.renderer.selection_text.clone();
-                    }
+
+                    histogram!(HistTag::Redraw).record(redraw_start.elapsed());
                 }
                 Event::WindowEvent { event, .. } => match event {
                     WindowEvent::Resized(size) => pending_resize = Some(size),
@@ -388,9 +393,7 @@ impl Inlyne {
                             .contains(position.into())
                                 && mouse_down)
                         {
-                            let scrollbar_height = (screen_size.1
-                                / self.renderer.positioner.reserved_height)
-                                * screen_size.1;
+                            let scrollbar_height = self.renderer.scrollbar_height();
                             if scrollbar_held.is_none() {
                                 if Rect::new(
                                     (
@@ -426,14 +429,10 @@ impl Inlyne {
                                 * self.renderer.positioner.reserved_height;
                             self.renderer.set_scroll_y(target_scroll);
                             self.window.request_redraw();
-                        } else if let Some(selection) = &mut self.renderer.selection {
-                            if mouse_down {
-                                selection.1 = loc;
-                                selecting = true;
-                                self.window.request_redraw();
-                            }
+                        } else if mouse_down && self.selection.handle_drag(loc) {
+                            self.window.request_redraw();
                         }
-                        last_loc = loc;
+                        mouse_position = loc;
                     }
                     WindowEvent::MouseInput {
                         state,
@@ -441,120 +440,98 @@ impl Inlyne {
                         ..
                     } => match state {
                         ElementState::Pressed => {
-                            // Reset selection
-                            if self.renderer.selection.is_some() {
-                                self.renderer.selection = None;
+                            // Try to click a link
+                            let screen_size = self.renderer.screen_size();
+
+                            let y = mouse_position.1 - self.renderer.scroll_y;
+                            if Rect::new(
+                                (screen_size.0 - DEFAULT_MARGIN / 4., 0.),
+                                (DEFAULT_MARGIN / 4., screen_size.1),
+                            ).contains((mouse_position.0, y)) {
+                                let scrollbar_height = self.renderer.scrollbar_height();
+
+                                let target_scroll = ((y - scrollbar_height / 2.) / screen_size.1)
+                                    * self.renderer.positioner.reserved_height;
+
+                                self.renderer.set_scroll_y(target_scroll);
                                 self.window.request_redraw();
                             }
 
-                            // Try to click a link
-                            let screen_size = self.renderer.screen_size();
                             if let Some(hoverable) = Self::find_hoverable(
                                 &mut self.renderer.text_system,
                                 &mut self.renderer.positioner.taffy,
                                 &self.elements,
-                                last_loc,
+                                mouse_position,
                                 screen_size,
                                 self.renderer.zoom,
                             ) {
-                                if let Hoverable::Summary(summary) = hoverable {
-                                    let mut hidden = summary.hidden.borrow_mut();
-                                    *hidden = !*hidden;
-                                    event_loop_proxy
-                                        .send_event(InlyneEvent::Reposition)
-                                        .unwrap();
-                                }
+                                match hoverable {
+                                    Hoverable::Image(Image { is_link: Some(link), .. }) |
+                                    Hoverable::Text(Text { link: Some(link), .. }) => {
+                                        let path = PathBuf::from_str(link).unwrap(); // Can't fail
 
-                                let maybe_link = match hoverable {
-                                    Hoverable::Image(Image { is_link, .. }) => is_link,
-                                    Hoverable::Text(Text { link, .. }) => link,
-                                    Hoverable::Summary(_) => &None,
-                                };
-
-                                if let Some(link) = maybe_link {
-                                    let maybe_path = PathBuf::from_str(link).ok();
-                                    let is_local_md = maybe_path.as_ref().map_or(false, |p| {
-                                        p.extension().map_or(false, |ext| ext == "md")
-                                            && !p.to_str().map_or(false, |s| s.starts_with("http"))
-                                    });
-                                    if is_local_md {
-                                        // Open markdown files ourselves
-                                        let path = maybe_path.expect("not a path");
-                                        // Handle relative paths and make them
-                                        // absolute by prepending current
-                                        // parent
-                                        let path = if path.is_relative() {
-                                            // Simply canonicalizing it doesn't suffice and leads to "no such file or directory"
-                                            let current_parent = self
-                                                .opts
-                                                .file_path
-                                                .parent()
-                                                .expect("no current parent");
-                                            let mut normalized_link = path.as_path();
-                                            if let Ok(stripped) = normalized_link
-                                                .strip_prefix(std::path::Component::CurDir)
-                                            {
-                                                normalized_link = stripped;
-                                            }
-                                            let mut link = current_parent.to_path_buf();
-                                            link.push(normalized_link);
-                                            link
-                                        } else {
-                                            path
-                                        };
-                                        // Open them in a new window, akin to what a browser does
-                                        if modifiers.shift() {
-                                            Command::new(
-                                                std::env::current_exe()
-                                                    .unwrap_or_else(|_| "inlyne".into()),
-                                            )
-                                            .args(Opts::program_args(&path))
-                                            .spawn()
-                                            .expect("Could not spawn new inlyne instance");
-                                        } else {
-                                            match read_to_string(&path) {
-                                                Ok(contents) => {
-                                                    self.opts.file_path = path;
-                                                    self.watcher.update_file(
-                                                        &self.opts.file_path,
-                                                        contents,
-                                                    );
-                                                    // TODO: Once and if history is implemented,
-                                                    // old scroll_y might be stored there
-                                                    self.renderer.set_scroll_y(0.);
-                                                }
-                                                Err(err) => {
-                                                    tracing::warn!(
+                                        if  path.extension().is_some_and(|ext| ext == "md")
+                                            && !path.to_str().is_some_and(|s| s.starts_with("http")) {
+                                            // Open them in a new window, akin to what a browser does
+                                            if modifiers.shift() {
+                                                std::thread::spawn(move || {
+                                                    Command::new(
+                                                        std::env::current_exe()
+                                                            .unwrap_or_else(|_| "inlyne".into()),
+                                                    )
+                                                        .args(Opts::program_args(&path))
+                                                        .spawn()
+                                                        .expect("Couldn't spawn inlyne instance")
+                                                        .wait()
+                                                        .expect("Failed waiting on child");
+                                                });
+                                            } else {
+                                                match read_to_string(&path) {
+                                                    Ok(contents) => {
+                                                        self.update_file(&path, contents);
+                                                        self.opts.history.make_next(path);
+                                                    }
+                                                    Err(err) => {
+                                                        tracing::warn!(
                                                         "Failed loading markdown file at {}\nError: {}",
                                                         path.display(),
                                                         err,
                                                     );
+                                                    }
                                                 }
                                             }
+                                        } else if let Some(anchor_pos) =
+                                            self.renderer.positioner.anchors.get(&link.to_lowercase())
+                                        {
+                                            self.renderer.set_scroll_y(*anchor_pos);
+                                            self.window.request_redraw();
+                                            self.window.set_cursor_icon(CursorIcon::Default);
+                                        } else if let Err(e) = open::that(link) {
+                                            tracing::error!("Could not open link: {e} from {:?}", std::env::current_dir())
                                         }
-                                    } else if let Some(anchor_pos) =
-                                        self.renderer.positioner.anchors.get(&link.to_lowercase())
-                                    {
-                                        self.renderer.set_scroll_y(*anchor_pos);
+                                    },
+                                    Hoverable::Summary(summary) => {
+                                        let mut hidden = summary.hidden.borrow_mut();
+                                        *hidden = !*hidden;
+                                        event_loop_proxy
+                                            .send_event(InlyneEvent::Reposition)
+                                            .unwrap();
+                                        self.selection.add_position(mouse_position);
+                                    },
+                                    _ => {
+                                        self.selection.add_position(mouse_position);
                                         self.window.request_redraw();
-                                        self.window.set_cursor_icon(CursorIcon::Default);
-                                    } else {
-                                        open::that(link).unwrap();
                                     }
-                                } else if self.renderer.selection.is_none() {
-                                    // Only set selection when not over link
-                                    self.renderer.selection = Some((last_loc, last_loc));
-                                }
-                            } else if self.renderer.selection.is_none() {
-                                self.renderer.selection = Some((last_loc, last_loc));
+                                };
+                            } else {
+                                self.selection.add_position(mouse_position);
+                                self.window.request_redraw()
                             }
-
                             mouse_down = true;
                         }
                         ElementState::Released => {
                             scrollbar_held = None;
                             mouse_down = false;
-                            selecting = false;
                         }
                     },
                     WindowEvent::ModifiersChanged(new_state) => modifiers = new_state,
@@ -624,8 +601,31 @@ impl Inlyne {
                                     self.window.request_redraw();
                                 }
                                 Action::Copy => clipboard
-                                    .set_contents(selection_cache.trim().to_owned()),
+                                    .set_contents(self.selection.text.trim().to_owned()),
                                 Action::Quit => *control_flow = ControlFlow::Exit,
+                                Action::History(hist_dir) => {
+                                    let changed_path = match hist_dir {
+                                        HistDirection::Next => self.opts.history.next(),
+                                        HistDirection::Prev => self.opts.history.previous(),
+                                    }.map(ToOwned::to_owned);
+                                    let Some(file_path) = changed_path else {
+                                        return;
+                                    };
+                                    match read_to_string(&file_path) {
+                                        Ok(contents) => {
+                                            self.update_file(&file_path, contents);
+                                            let parent = file_path.parent().expect("File should have parent directory");
+                                            std::env::set_current_dir(parent).expect("Could not set current directory.");
+                                        }
+                                        Err(err) => {
+                                            tracing::warn!(
+                                                "Failed loading markdown file at {}\nError: {}",
+                                                file_path.display(),
+                                                err,
+                                            );
+                                        }
+                                    }
+                                }
                             }
                         }
                     }
@@ -634,7 +634,7 @@ impl Inlyne {
                 Event::MainEventsCleared => {
                     // We lazily store the size and only reposition elements and request a redraw when
                     // we receive a `MainEventsCleared`.  This prevents us from clogging up the queue
-                    // with a bunch of costly resizes. (https://github.com/trimental/inlyne/issues/25)
+                    // with a bunch of costly resizes. (https://github.com/Inlyne-Project/inlyne/issues/25)
                     if let Some(size) = pending_resize.take() {
                         if size.width > 0 && size.height > 0 {
                             self.renderer.config.width = size.width;
@@ -754,7 +754,7 @@ impl Inlyne {
 }
 
 fn main() -> anyhow::Result<()> {
-    human_panic::setup_panic!();
+    setup_panic!();
 
     let env_filter = tracing_subscriber::EnvFilter::builder()
         .with_default_directive("inlyne=info".parse()?)
@@ -765,21 +765,65 @@ fn main() -> anyhow::Result<()> {
         .with(tracing_subscriber::fmt::layer().compact())
         .init();
 
-    let args = Args::new();
-    let config = match &args.config {
-        Some(config_path) => Config::load_from_file(config_path)?,
-        None => Config::load_from_system().unwrap_or_else(|err| {
-            tracing::warn!(
-                "Failed reading config file. Falling back to defaults. Error: {}",
-                err
-            );
-            Config::default()
-        }),
-    };
-    let opts = Opts::parse_and_load_from(args, config)?;
+    let command = Cli::parse().into_commands();
 
-    let inlyne = Inlyne::new(opts)?;
-    inlyne.run();
+    match command {
+        Commands::View(view) => {
+            let config = match &view.config {
+                Some(config_path) => Config::load_from_file(config_path)?,
+                None => Config::load_from_system().unwrap_or_else(|err| {
+                    tracing::warn!(
+                        "Failed reading config file. Falling back to defaults. Error: {}",
+                        err
+                    );
+                    Config::default()
+                }),
+            };
+            let opts = Opts::parse_and_load_from(view, config)?;
+
+            if let Some(exporter) = &opts.metrics {
+                match exporter {
+                    MetricsExporter::Log => {
+                        let recorder = metrics::LogRecorder::default();
+                        metrics::set_global_recorder(recorder)
+                            .expect("Failed setting metrics recorder");
+                    }
+                    #[cfg(inlyne_tcp_metrics)]
+                    MetricsExporter::Tcp => metrics_exporter_tcp::TcpBuilder::new()
+                        .install()
+                        .expect("Failed to install TCP metrics server"),
+                };
+            }
+
+            for tag in HistTag::iter() {
+                tag.set_global_description();
+            }
+
+            let inlyne = Inlyne::new(opts)?;
+            inlyne.run();
+        }
+        Commands::Config(ConfigCmd::Open) => {
+            let config_path = dirs::config_dir()
+                .context("Failed to find the configuration directory")?
+                .join("inlyne")
+                .join("inlyne.toml");
+
+            let config = std::fs::read_to_string(&config_path)
+                .unwrap_or_else(|_| Config::default_config().to_string());
+
+            let new_config = edit::edit_with_builder(
+                &config,
+                edit::Builder::new()
+                    .prefix("inlyne_temp")
+                    .suffix(".toml")
+                    .keep(true),
+            )?;
+
+            _ = Config::load_from_str(&new_config)?;
+
+            std::fs::write(config_path, new_config)?;
+        }
+    }
 
     Ok(())
 }
