@@ -1,14 +1,17 @@
 use std::borrow::Cow;
 use std::sync::{Arc, Mutex};
+use std::time::Instant;
 
 use crate::color::{native_color, Theme};
 use crate::fonts::get_fonts;
 use crate::image::ImageRenderer;
+use crate::metrics::{histogram, HistTag};
 use crate::opts::FontOptions;
 use crate::positioner::{Positioned, Positioner, DEFAULT_MARGIN};
+use crate::selection::Selection;
 use crate::table::TABLE_ROW_GAP;
 use crate::text::{CachedTextArea, TextCache, TextSystem};
-use crate::utils::{Point, Rect, Selection, Size};
+use crate::utils::{Point, Rect, Size};
 use crate::Element;
 
 use anyhow::{Context, Ok};
@@ -43,8 +46,6 @@ pub struct Renderer {
     pub page_width: f32,
     pub image_renderer: ImageRenderer,
     pub theme: Theme,
-    pub selection: Option<Selection>,
-    pub selection_text: String,
     pub zoom: f32,
     pub positioner: Positioner,
 }
@@ -184,14 +185,15 @@ impl Renderer {
             zoom: 1.,
             image_renderer,
             theme,
-            selection: None,
-            selection_text: String::new(),
             positioner,
         })
     }
 
     fn draw_scrollbar(&mut self) -> anyhow::Result<()> {
         let (screen_width, screen_height) = self.screen_size();
+        if screen_height > self.positioner.reserved_height {
+            return Ok(());
+        }
         let height = (screen_height / self.positioner.reserved_height) * screen_height;
         self.draw_rectangle(
             Rect::new(
@@ -206,9 +208,14 @@ impl Renderer {
         Ok(())
     }
 
+    pub fn scrollbar_height(&self) -> f32 {
+        (self.screen_height() / self.positioner.reserved_height) * self.screen_height()
+    }
+
     fn render_elements(
         &mut self,
         elements: &[Positioned<Element>],
+        selection: &mut Selection,
     ) -> anyhow::Result<Vec<CachedTextArea>> {
         let mut text_areas: Vec<CachedTextArea> = Vec::new();
         let screen_size = self.screen_size();
@@ -296,13 +303,14 @@ impl Renderer {
                         }
                     }
                     if let Some(is_checked) = text_box.is_checkbox {
+                        let line_height = text_box.line_height(self.zoom);
                         let min = (
                             scrolled_pos.0 - box_size * 1.5,
-                            scrolled_pos.1 + size.1 / 2. - box_size / 2.,
+                            scrolled_pos.1 + line_height / 2. - box_size / 2.,
                         );
                         let max = (
                             scrolled_pos.0 + box_size - box_size * 1.5,
-                            scrolled_pos.1 + size.1 / 2. + box_size / 2.,
+                            scrolled_pos.1 + line_height / 2. + box_size / 2.,
                         );
                         if max.0 < screen_size.0 - DEFAULT_MARGIN - centering {
                             if is_checked {
@@ -335,16 +343,13 @@ impl Renderer {
                         let max = (line.max.0, line.max.1 + 2. * self.hidpi_scale * self.zoom);
                         self.draw_rectangle(Rect::from_min_max(min, max), line.color)?;
                     }
-                    if let Some(selection) = self.selection {
-                        let (selection_rects, selection_text) = text_box.render_selection(
-                            &mut self.text_system,
-                            pos,
-                            bounds,
-                            self.zoom,
-                            selection,
-                        );
-                        self.selection_text.push_str(&selection_text);
-                        self.selection_text.push('\n');
+                    if let Some(selection_rects) = text_box.render_selection(
+                        &mut self.text_system,
+                        pos,
+                        bounds,
+                        self.zoom,
+                        selection,
+                    ) {
                         for rect in selection_rects {
                             self.draw_rectangle(
                                 Rect::from_min_max(
@@ -377,16 +382,13 @@ impl Renderer {
                                 self.zoom,
                                 self.scroll_y,
                             ));
-                            if let Some(selection) = self.selection {
-                                let (selection_rects, selection_text) = text_box.render_selection(
-                                    &mut self.text_system,
-                                    (pos.0 + node.location.x, pos.1 + node.location.y),
-                                    (node.size.width, node.size.height),
-                                    self.zoom,
-                                    selection,
-                                );
-                                self.selection_text.push_str(&selection_text);
-                                self.selection_text.push('\n');
+                            if let Some(selection_rects) = text_box.render_selection(
+                                &mut self.text_system,
+                                (pos.0 + node.location.x, pos.1 + node.location.y),
+                                (node.size.width, node.size.height),
+                                self.zoom,
+                                selection,
+                            ) {
                                 for rect in selection_rects {
                                     self.draw_rectangle(
                                         Rect::from_min_max(
@@ -440,17 +442,13 @@ impl Renderer {
                                         self.scroll_y,
                                     ));
 
-                                    if let Some(selection) = self.selection {
-                                        let (selection_rects, selection_text) = text_box
-                                            .render_selection(
-                                                &mut self.text_system,
-                                                (pos.0 + node.location.x, pos.1 + node.location.y),
-                                                (node.size.width, node.size.height),
-                                                self.zoom,
-                                                selection,
-                                            );
-                                        self.selection_text.push_str(&selection_text);
-                                        self.selection_text.push('\n');
+                                    if let Some(selection_rects) = text_box.render_selection(
+                                        &mut self.text_system,
+                                        (pos.0 + node.location.x, pos.1 + node.location.y),
+                                        (node.size.width, node.size.height),
+                                        self.zoom,
+                                        selection,
+                                    ) {
                                         for rect in selection_rects {
                                             self.draw_rectangle(
                                                 Rect::from_min_max(
@@ -510,7 +508,9 @@ impl Renderer {
                         )?;
                     }
                 }
-                Element::Row(row) => text_areas.append(&mut self.render_elements(&row.elements)?),
+                Element::Row(row) => {
+                    text_areas.append(&mut self.render_elements(&row.elements, selection)?)
+                }
                 Element::Section(section) => {
                     if let Some(ref summary) = *section.summary {
                         let bounds = summary.bounds.as_ref().unwrap();
@@ -523,15 +523,27 @@ impl Renderer {
                             native_color(self.theme.text_color, &self.surface_format),
                             *section.hidden.borrow(),
                         )?;
-                        text_areas.append(&mut self.render_elements(std::slice::from_ref(summary))?)
+                        text_areas.append(
+                            &mut self.render_elements(std::slice::from_ref(summary), selection)?,
+                        )
                     }
                     if !*section.hidden.borrow() {
-                        text_areas.append(&mut self.render_elements(&section.elements)?)
+                        text_areas.append(&mut self.render_elements(&section.elements, selection)?)
                     }
                 }
             }
-        }
 
+            if crate::opts::get_render_element_bounds() {
+                let mut rect = element
+                    .bounds
+                    .as_ref()
+                    .context("Element not positioned")?
+                    .clone();
+                rect.pos.1 -= self.scroll_y;
+                let color = glyphon::Color::rgb(255, 0, 255).0;
+                let _ = self.stroke_rectangle(rect, native_color(color, &self.surface_format), 1.0);
+            }
+        }
         self.draw_scrollbar()?;
         Ok(text_areas)
     }
@@ -719,7 +731,12 @@ impl Renderer {
         bind_groups
     }
 
-    pub fn redraw(&mut self, elements: &mut [Positioned<Element>]) -> anyhow::Result<()> {
+    pub fn redraw(
+        &mut self,
+        elements: &mut [Positioned<Element>],
+        selection: &mut Selection,
+    ) -> anyhow::Result<()> {
+        selection.text.clear();
         let frame = self
             .surface
             .get_current_texture()
@@ -734,8 +751,7 @@ impl Renderer {
         // Prepare and render elements that use lyon
         self.lyon_buffer.indices.clear();
         self.lyon_buffer.vertices.clear();
-        self.selection_text = String::new();
-        let cached_text_areas = self.render_elements(elements)?;
+        let cached_text_areas = self.render_elements(elements, selection)?;
         let vertex_buf = self
             .device
             .create_buffer_init(&wgpu::util::BufferInitDescriptor {
@@ -828,8 +844,12 @@ impl Renderer {
     }
 
     pub fn reposition(&mut self, elements: &mut [Positioned<Element>]) -> anyhow::Result<()> {
-        self.positioner
-            .reposition(&mut self.text_system, elements, self.zoom)
+        let start = Instant::now();
+        let res = self
+            .positioner
+            .reposition(&mut self.text_system, elements, self.zoom);
+        histogram!(HistTag::Reposition).record(start.elapsed());
+        res
     }
 
     pub fn set_scroll_y(&mut self, scroll_y: f32) {
