@@ -2,6 +2,7 @@
 
 include!("../../src/opts/cli.rs");
 
+use clap::CommandFactory;
 use std::fs;
 
 use clap_complete::{generate_to, Shell};
@@ -14,7 +15,7 @@ fn gen_completions() {
     let out_dir = "completions";
     fs::create_dir_all(out_dir).unwrap();
 
-    let mut cmd = command();
+    let mut cmd = Cli::command();
     for &shell in Shell::value_variants() {
         generate_to(shell, &mut cmd, "inlyne", out_dir).unwrap();
     }
